#include "Classif.h"

// chargement des donnees
void Classif::loadData(char* name,const string& typesAttributs) {
	ifstream countLinesFile(name);
	if (!countLinesFile.is_open()) {
		cout << "erreur, le fichier " << name << " n'existe pas" << endl;
		exit(1);
	}
	// on cherche d'abord le nombre de lignes de la BDD
	nbLines=0; 
	while (countLinesFile.ignore(numeric_limits<int>::max(),'\n')) {
		nbLines++;
	}
	countLinesFile.close();

	// initialisation des structures de donnees (une colonne~=un pointeur)
	int* tabIndex=new int[typesAttributs.size()]; // indique les index des types d'attributs dans leurs tableaux respectifs
	int nbDisc=0,nbCont=0,nbOrd=0; // pour le remplissage du tableau d'index
	for (int i=0; i<=(int)typesAttributs.size(); i++) {
		if (typesAttributs[i]=='d') {
			attDiscrets.push_back(new StringOuMiss[nbLines]);
			tabIndex[i]=nbDisc;
			nbDisc++;
		}
		else if (typesAttributs[i]=='c') {
			attContinus.push_back(new FloatOuMiss[nbLines]);
			tabIndex[i]=nbCont;
			nbCont++;
		}
		else if (typesAttributs[i]=='o') {
			attDiscOrdonnes.push_back(new IntOuMiss[nbLines]);
			tabIndex[i]=nbOrd;
			nbOrd++;
		}
	}

	// on recopie ensuite la base de donnees dans une structure plus adequate
	ifstream dataFile(name);
	string line,nomClasse,strTmp;
	nbClasses=0;
	classes=new string[nbLines];
	int countLines=0,countAtt,i;
	vector<StringIndexCounter> classesDejaVues; // les noms des classes deja rencontrees
	string::iterator curChar,startPoint,memCurChar;
	bool newSubstring; // indique si on est en train de lire une nouvelle valeur
	while (getline(dataFile,line,'\n')) {
		if (line.empty()) break; // escaper si \n en fin de fichier
		countAtt=0; // compte le nombre d'attributs lus
		newSubstring=true;
		for (curChar=line.begin(); curChar!=line.end(); curChar++) {
			if (newSubstring && *curChar!=',') {
				startPoint=curChar;
				newSubstring=false;
			}
			else if (*curChar==',') { // on ignore les espaces au debut et a la fin des chaines
				while (*startPoint==' ') startPoint++;
				memCurChar=curChar;
				while (*(curChar-1)==' ') curChar--;
				strTmp=string(startPoint,curChar);
				curChar=memCurChar;
				if (strTmp=="?") {
					if (typesAttributs[countAtt]=='d')
						attDiscrets[tabIndex[countAtt]][countLines].miss=true;
					else if (typesAttributs[countAtt]=='c') {
						attContinus[tabIndex[countAtt]][countLines].miss=true;
						attContinus[tabIndex[countAtt]][countLines].isInit=true;
					}
					else if (typesAttributs[countAtt]=='o') {
						attDiscOrdonnes[tabIndex[countAtt]][countLines].miss=true;
						attDiscOrdonnes[tabIndex[countAtt]][countLines].isInit=true;
					}
				}
				else {
					if (typesAttributs[countAtt]=='d')
						attDiscrets[tabIndex[countAtt]][countLines].str=strTmp;
					else if (typesAttributs[countAtt]=='c') {
						attContinus[tabIndex[countAtt]][countLines].val=toFloat(strTmp);
						attContinus[tabIndex[countAtt]][countLines].isInit=true;
					}
					else if (typesAttributs[countAtt]=='o') {
						attDiscOrdonnes[tabIndex[countAtt]][countLines].val=toInt(strTmp);
						attDiscOrdonnes[tabIndex[countAtt]][countLines].isInit=true;
					}
				}
				newSubstring=true;
				countAtt++;
			}
			if (countAtt==(int)typesAttributs.size()) break;
		}
		startPoint=++curChar;
		while (curChar!=line.end()) curChar++;
		while (*startPoint==' ') startPoint++;
		while (*(curChar-1)==' ') curChar--;
		nomClasse=string(startPoint,curChar);
		classes[countLines]=nomClasse;
		for (i=classesDejaVues.size()-1; i>=0; i--) {
			if (nomClasse==classesDejaVues[i].str) {
				classesDejaVues[i].counter++;
				classesDejaVues[i].occurrences.push_back(countLines);
				break;
			}
		}
		if (i<0) {
			classesDejaVues.push_back(StringIndexCounter(nomClasse,nbClasses,1));
			classesDejaVues.back().occurrences.push_back(countLines);
			nbClasses++;
		}
		countLines++;
	}
	dataFile.close();
	delete [] tabIndex;
	abr=Arbre(classesDejaVues);

	// remplissage du tableau des puissances de 2 necessaires a l'acceleration des calculs
	puiss2=new unsigned long long[nbClasses];
	for (i=nbClasses-1; i>=0; i--) puiss2[i]=1;
	for (i=1; i<nbClasses; i++) puiss2[i]=2*puiss2[i-1];

	// test pour savoir si nbLines ou nbLines-1 ..etc (si plusieurs sauts de ligne en fin de fichier)
	// on tire parti du fait que les chaines sont initialisees a "" (similaires aux vecteurs)
	i=nbLines-1;
	if (!attDiscrets.empty())
		while (attDiscrets[0][i].str.empty()) i--;
	else if (!attContinus.empty())
		while (!attContinus[0][i].isInit) i--;
	else
		while (!attDiscOrdonnes[0][i].isInit) i--;
	nbLines=i+1; // important pour les stats finales
}	
	
// la fonction de creation de l'arbre de classification
void Classif::buildTree(char spC) {
	int curID=0;
	splitCriteria=spC;
	abr.racine->calculateSplitValue(splitCriteria);
	Noeud *node,**newFils;
	vector<VectorString> subsets;
	float cutFloat,valSplitDisc,valSplitCont,valSplitOrd;
	int cutInt,indexAttString,indexAttFloat,indexAttInt,index,nbFilsNV;
	char maximum;
	while (abr.getOpenNode(&node)) { // on choisit arbitrairement un noeud de l'arbre non resolu
									// le critere d'arret pourra [devra..] etre assoupli plus tard
		node->id=curID++;
		valSplitDisc=findBestSplitDiscrete(node,subsets,indexAttString);
		valSplitCont=findBestSplitContinue(node,cutFloat,indexAttFloat);
		valSplitOrd=findBestSplitOrdered(node,cutInt,indexAttInt);
		// dans le cas ou on n'a pas pu ameliorer le split indice, on continue sans engendrer de fils
		if (valSplitDisc<0.0 && valSplitCont<0.0 && valSplitOrd<0.0) {
			node->type='f';
			node->setLabel();
			continue;
		}
		maximum=max(valSplitDisc,valSplitCont,valSplitOrd);

		// les missing_values sont prises en compte lors de la recherche du meilleur split, comme vu en cours
		// on separe les donnees:
		if (maximum=='d') {
			node->type='d'; // jusqu'ici le type n'etait pas connu
			node->indexAtt=indexAttString;
			node->subsets=subsets;
			newFils=new Noeud*[subsets.size()];
			for (int i=subsets.size()-1; i>=0; i--) newFils[i]=new Noeud();
			
			// parcours des exemples du noeud pere pour les distribuer dans les noeuds fils;
			// les missing_value sont ajoutes aux fils avec une distribution de probabilite
			// proportionnelle aux cardinaux des differentes classes presentes (sauf si trop peu significatif)
			for (int i=node->distribExemples.size()-1; i>=0; i--) {
				for (int j=node->distribExemples[i].occurrences.size()-1; j>=0; j--) {
					if (!attDiscrets[indexAttString][node->distribExemples[i].occurrences[j]].miss) {
						newFils[getIndexIn(subsets,attDiscrets[indexAttString][node->distribExemples[i].occurrences[j]].str)]->
							addInStringIndexCounter(node->distribExemples[i].str,node->distribExemples[i].index,node->distribExemples[i].occurrences[j]);
					}
					else {
						// valeur manquante
						index=findIndexDiscrete(node,subsets,indexAttString,classes[node->distribExemples[i].occurrences[j]]);
						if (index>=0)
							newFils[index]->addInStringIndexCounter(node->distribExemples[i].str,node->distribExemples[i].index,node->distribExemples[i].occurrences[j]);
					}
				}
			}
			// ajout des fils au noeud courant, eventuellement calcul des split indices de base + link sur pere
			node->addSuccessors(newFils,subsets.size());
		}
		else {
			node->type=maximum; // jusqu'ici le type n'etait pas connu
			if (maximum=='c') {
				node->indexAtt=indexAttFloat;
				node->cutFloat=cutFloat;
			}
			else {
				node->indexAtt=indexAttInt;
				node->cutInt=cutInt;
			}
			newFils=new Noeud*[2];
			newFils[0]=new Noeud();
			newFils[1]=new Noeud();
			// parcours des exemples du noeud pere pour les distribuer dans les noeuds fils;
			// les missing_value sont ajoutes aux fils avec une distribution de probabilite
			// proportionnelle aux cardinaux des differentes classes presentes (sauf si trop peu significatif)
			for (int i=node->distribExemples.size()-1; i>=0; i--) {
				for (int j=node->distribExemples[i].occurrences.size()-1; j>=0; j--) {
					if (maximum=='c') {
						if (!attContinus[indexAttFloat][node->distribExemples[i].occurrences[j]].miss) {
							if (attContinus[indexAttFloat][node->distribExemples[i].occurrences[j]].val<=cutFloat)
								newFils[0]->addInStringIndexCounter(node->distribExemples[i].str,node->distribExemples[i].index,node->distribExemples[i].occurrences[j]);
							else 
								newFils[1]->addInStringIndexCounter(node->distribExemples[i].str,node->distribExemples[i].index,node->distribExemples[i].occurrences[j]);
						}
						else {
							// valeur manquante
							index=findIndexContinue(node,cutFloat,indexAttFloat,classes[node->distribExemples[i].occurrences[j]]);
							if (index>=0)
								newFils[index]->addInStringIndexCounter(node->distribExemples[i].str,node->distribExemples[i].index,node->distribExemples[i].occurrences[j]);
						}
					}
					else {
						if (!attDiscOrdonnes[indexAttInt][node->distribExemples[i].occurrences[j]].miss) {
							if (attDiscOrdonnes[indexAttInt][node->distribExemples[i].occurrences[j]].val<=cutInt)
								newFils[0]->addInStringIndexCounter(node->distribExemples[i].str,node->distribExemples[i].index,node->distribExemples[i].occurrences[j]);
							else 
								newFils[1]->addInStringIndexCounter(node->distribExemples[i].str,node->distribExemples[i].index,node->distribExemples[i].occurrences[j]);
						}
						else {
							// valeur manquante
							index=findIndexOrdered(node,cutInt,indexAttInt,classes[node->distribExemples[i].occurrences[j]]);
							if (index>=0)
								newFils[index]->addInStringIndexCounter(node->distribExemples[i].str,node->distribExemples[i].index,node->distribExemples[i].occurrences[j]);
						}
					}
				}
			}
			// ajout des fils au noeud courant, eventuellement calcul des split indices de base + link sur pere
			node->addSuccessors(newFils,2);
		}
		// apres split, on met a jour les noeuds ouverts (fils du noeud juste traite)
		// s'il n'y a qu'un fils "non vide", on s'arrete
		nbFilsNV=0;
		index=0;
		for (int i=node->nbFils-1; i>=0; i--) {
			newFils[i]->setFather(node); // pointeur sur pere
			newFils[i]->calculateSplitValue(splitCriteria);
			if (newFils[i]->splitValue>0.0) {
				abr.openNodes.push_back(newFils[i]);
				nbFilsNV++;
				index=i;
			}
			else {
				newFils[i]->id=curID++;
				if (!(newFils[i]->distribExemples.empty())) nbFilsNV++;
				newFils[i]->type='f';
				newFils[i]->setLabel();
			}
		}
		if (nbFilsNV<=1) {
			abr.openNodes.pop_back();
			newFils[index]->type='f';
			newFils[index]->setLabel();
		}
	}
}

// des fonctions auxiliaires qui retourne la meilleure valeur de split
// retourne -1 si type d'attribut non present
// on utilise la methode "reduction d'entropie/Gini/misc_error" (divisee par split_ratio dans le cas discret symbolique)
float Classif::findBestSplitDiscrete(Noeud* node,vector<VectorString>& subsets,int& indexAtt) {
	if (attDiscrets.empty()) return -1.0;
	vector<VectorString> curSubsets;
	vector<StringAndClassesID> chainesClasses; // chaines (symboles) rencontrees au temps + classes correspondantes
	float bestSplitVal=-1.0,proportion,maxProp;
	float* proportions=new float[nbClasses];
	int countMissingValues; // valeurs manquantes
	// methode: on split sur toutes les valeurs discretes puis on fait des regroupements eventuels
	for (int i=attDiscrets.size()-1; i>=0; i--) {
		// tentative pour l'attribut symbolique numero i

		// recherche de tous les symboles presents pour l'attribut i: un vecteur par symbole pour commencer
		curSubsets.clear();
		chainesClasses.clear();
		countMissingValues=0;
		for (int j=node->distribExemples.size()-1; j>=0; j--) {
			for (int k=node->distribExemples[j].occurrences.size()-1; k>=0; k--) {
				if (!attDiscrets[i][node->distribExemples[j].occurrences[k]].miss) {
					int l;
					for (l=chainesClasses.size()-1; l>=0; l--) 
						if (attDiscrets[i][node->distribExemples[j].occurrences[k]].str==chainesClasses[l].str) break;
					if (l>=0) continue;
					VectorString vs;
					vs.push_back(attDiscrets[i][node->distribExemples[j].occurrences[k]].str);
					curSubsets.push_back(vs);
					unsigned long long classesID=0;
					bool* pbool=(bool*)calloc(nbClasses,sizeof(bool));
					chainesClasses.push_back(StringAndClassesID(attDiscrets[i][node->distribExemples[j].occurrences[k]].str,classesID,pbool));
				}
				else countMissingValues++;
			}
		}

		// on regarde si des regroupements sont possibles:
		// on associe un "vecteur de bits" a chaque chaine trouvee, representant les classes qui ont cette valeur d'attribut.
		// dans le cas ou deux chaines ont le meme vecteur de bits, il faut les fusionner;
		// pour l'instant c'est le seul cas de fusion que l'on envisage (sinon complexite exponentielle)
		for (int j=node->distribExemples.size()-1; j>=0; j--) {
			for (int k=node->distribExemples[j].occurrences.size()-1; k>=0; k--) {
				// si la valeur de l'attribut est manquante, on continue:
				if (attDiscrets[i][node->distribExemples[j].occurrences[k]].miss) continue;
				int l;
				for (l=chainesClasses.size()-1; l>=0; l--) 
					if (chainesClasses[l].str==attDiscrets[i][node->distribExemples[j].occurrences[k]].str) break;
				if (!chainesClasses[l].pbool[node->distribExemples[j].index]) {
					chainesClasses[l].classesID+=puiss2[node->distribExemples[j].index];
					chainesClasses[l].pbool[node->distribExemples[j].index]=true;
				}
			}
		}

		// plus qu'a apparier les chaines aux classesID similaires, SAUF si un seul classesID (==on ne ferait rien en regroupant tout!)
		unsigned long long valTest=chainesClasses.back().classesID; // necessairement valTest>0 dans au moins une classe
		for (int j=chainesClasses.size()-2; j>=0; j--) {
			if (chainesClasses[j].classesID!=valTest) {
				valTest=0;
				break;
			}
		}
		if (valTest<=0) {
			bool* chaineDejaAppariee=(bool*)calloc(chainesClasses.size(),sizeof(bool)); // memorisation des chaines traitees
			for (int j=0; j<(int)chainesClasses.size(); j++) {
				for (int k=j+1; k<(int)chainesClasses.size(); k++) {
					if (!chaineDejaAppariee[k]) {
						if (chainesClasses[k].classesID==chainesClasses[j].classesID) {
							curSubsets[j].push_back(chainesClasses[k].str);
							curSubsets[k].pop_back();
							chaineDejaAppariee[k]=true;
						}
					}
				}
			}
			free(chaineDejaAppariee);

			// on supprime les cellules vides en fin du vecteur curSubsets
			for (int j=0; j<(int)curSubsets.size(); j++) {
				if (curSubsets[j].empty()) {
					// on echange avec le premier sous-ensemble non vide que l'on trouve a partir de la (s'il en existe..)
					int k;
					for (k=j+1; k<(int)curSubsets.size(); k++)
						if (!curSubsets[k].empty()) break;
					if (k>=(int)curSubsets.size()) break;
					while (!curSubsets[k].empty()) {
						curSubsets[j].push_back(curSubsets[k].back());
						curSubsets[k].pop_back();
					}
				}
			}
			while (!curSubsets.empty() && curSubsets.back().empty()) curSubsets.pop_back();
		}

		// calcul de l'indice de discrimination avec curSubsets (attention, gain relatif d'information, on maximise!)
		int nbTotalExemples=0;
		int* nbExemples=(int*)calloc(curSubsets.size(),sizeof(int));
		float* splitIndice=new float[curSubsets.size()];
		for (int j=curSubsets.size()-1; j>=0; j--) {
			// pour chaque sous-ensemble il faut calculer le split indice, puis mixer le tout comme d'hab
			for (int k=nbClasses-1; k>=0; k--) proportions[k]=0;
			for (int k=node->distribExemples.size()-1; k>=0; k--) {
				for (int l=node->distribExemples[k].occurrences.size()-1; l>=0; l--) {
					if (isIn(attDiscrets[i][node->distribExemples[k].occurrences[l]].str,curSubsets[j])) {
						proportions[node->distribExemples[k].index]+=1.0;
						nbExemples[j]++;
						nbTotalExemples++;
					}
				}
			}
			// calcul de splitIndice[j]
			if (nbExemples[j]==0) {
				splitIndice[j]=0.0;
				continue;
			}
			switch (splitCriteria) {
				case 'g':
					splitIndice[j]=1.0;
					for (int k=nbClasses-1; k>=0; k--) splitIndice[j]-=Noeud::square(proportions[k]/nbExemples[j]);
					break;
				case 'e':
					splitIndice[j]=0.0;
					for (int k=nbClasses-1; k>=0; k--) {
						proportion=proportions[k]/nbExemples[j];
						if (proportion>0) splitIndice[j]-=proportion*log(proportion);
					}
					break;
				case 'm':
					splitIndice[j]=1.0;
					maxProp=0.0;
					for (int k=nbClasses-1; k>=0; k--) {
						proportion=proportions[k]/nbExemples[j];
						if (proportion>maxProp) maxProp=proportion;
					}
					splitIndice[j]-=maxProp;
					break;
				default: 
					splitIndice[j]=0.0; // jamais atteint mais evite les warnings
					break;
			}
		}
		// calcul de splitVal avec les splitIndice[j] et les nbExemples[j] + nombre total d'exemples
		// encore une fois, attention a ne pas diviser par 0
		float splitVal,tmpVal,branchFact;
		if (splitCriteria!='m') splitVal=node->splitValue;
		else splitVal=1.0; // car trop souvent la valeur recomposee est tres proche de celle du pere
		tmpVal=0.0;
		for (int j=curSubsets.size()-1; j>=0; j--) tmpVal+=nbExemples[j]*splitIndice[j];
		tmpVal/=(nbTotalExemples+countMissingValues);
		
		// on poursuit seulement si amelioration par rapport au split indice du pere
		if (tmpVal<node->splitValue) {
			splitVal-=tmpVal; // ici on est sur que nbTotalExemples>0, car node dans open
			// normalisation par le facteur de branchement:
			branchFact=0.0;
			for (int j=curSubsets.size()-1; j>=0; j--) {
				proportion=(float)nbExemples[j]/(nbTotalExemples+countMissingValues);
				if (proportion>0) branchFact-=proportion*log(proportion);
			}
			if (branchFact>0.0) splitVal/=branchFact;
			else splitVal-=1.0; // on ne veut pas des attributs non discriminants!
			if (splitVal>bestSplitVal) {
				bestSplitVal=splitVal;
				subsets=curSubsets;
				indexAtt=i;
			}
		}

		// liberation memoire
		free(nbExemples);
		delete [] splitIndice;
	}
	
	delete [] proportions;
	return bestSplitVal;
}
float Classif::findBestSplitContinue(Noeud* node,float& cutFloat,int& indexAtt) {
	if (attContinus.empty()) return -1.0;
	// on commence par trier les attributs par valeurs croissantes,
	// puis on teste de l'ordre de sqrt(n) splits differents equirepartis
	// on garde le meilleur (tous attributs continus confondus), tenant compte des missing_values
	float bestSplitVal=-1.0;
	vector<FloatAndClass> valeursAtt;
	float* proportions=new float[nbClasses];
	for (int i=attContinus.size()-1; i>=0; i--) {
		// tentative pour l'attribut continu numero i

		// a) recuperation des valeurs des attributs
		int countMissingValues=0;
		valeursAtt.clear();
		for (int j=node->distribExemples.size()-1; j>=0; j--) {
			for (int k=node->distribExemples[j].occurrences.size()-1; k>=0; k--) {
				if (!attContinus[i][node->distribExemples[j].occurrences[k]].miss)
					valeursAtt.push_back(FloatAndClass(attContinus[i][node->distribExemples[j].occurrences[k]].val,node->distribExemples[j].index));
				else countMissingValues++;
			}
		}
		// tri des valeurs
		sort(valeursAtt.begin(),valeursAtt.end());

		// b) tests de splits
		float range=valeursAtt.back().val-valeursAtt[0].val; // amplitude de variation de l'attribut
		float step=range/sqrt((float)valeursAtt.size()); // ecart entre deux tests de splits
		float maxProp,proportion;
		for (float f=valeursAtt[0].val+step/2.0; f<valeursAtt.back().val; f+=step) {

			// calcul du split indice a gauche: on cherche d'abord les proportions
			for (int j=nbClasses-1; j>=0; j--) proportions[j]=0.0; // remise a zero des proportions
			int nbTotalExemples=0,k;
			for (k=0;;k++) {
				if (valeursAtt[k].val>f) break;
				nbTotalExemples++;
				proportions[valeursAtt[k].classIndex]+=1.0;
			}
			float splitGauche;
			if (nbTotalExemples==0) splitGauche=0.0; // attention aux divisions par 0!
			else {
				switch (splitCriteria) {
					case 'g':
						splitGauche=1.0;
						for (int j=nbClasses-1; j>=0; j--) splitGauche-=Noeud::square(proportions[j]/nbTotalExemples);
						break;
					case 'e':
						splitGauche=0.0;
						for (int j=nbClasses-1; j>=0; j--) {
							proportion=proportions[j]/nbTotalExemples;
							if (proportion>0) splitGauche-=proportion*log(proportion);
						}
						break;
					case 'm':
						splitGauche=1.0;
						maxProp=0.0;
						for (int j=nbClasses-1; j>=0; j--) {
							proportion=proportions[j]/nbTotalExemples;
							if (proportion>maxProp) maxProp=proportion;
						}
						splitGauche-=maxProp;
						break;
					default: 
						splitGauche=0.0; // jamais atteint mais evite les warnings
						break;
				}
			}

			// calcul du split indice a droite: on cherche d'abord les proportions
			for (int j=nbClasses-1; j>=0; j--) proportions[j]=0.0; // remise a zero des proportions
			nbTotalExemples=0;
			for (;k<(int)valeursAtt.size();k++) {
				nbTotalExemples++;
				proportions[valeursAtt[k].classIndex]+=1.0;
			}
			float splitDroite;
			if (nbTotalExemples==0) splitDroite=0.0; // attention aux divisions par 0!
			else {
				switch (splitCriteria) {
					case 'g':
						splitDroite=1.0;
						for (int j=nbClasses-1; j>=0; j--) splitDroite-=Noeud::square(proportions[j]/nbTotalExemples);
						break;
					case 'e':
						splitDroite=0.0;
						for (int j=nbClasses-1; j>=0; j--) {
							proportion=proportions[j]/nbTotalExemples;
							if (proportion>0) splitDroite-=proportion*log(proportion);
						}
						break;
					case 'm':
						splitDroite=1.0;
						maxProp=0.0;
						for (int j=nbClasses-1; j>=0; j--) {
							proportion=proportions[j]/nbTotalExemples;
							if (proportion>maxProp) maxProp=proportion;
						}
						splitDroite-=maxProp;
						break;
					default: 
						splitDroite=0.0; // jamais atteint mais evite les warnings
						break;
				}
			}

			// on peut alors donner le split indice pour la valeur f:
			// normalisation (ameliore les perfs de l'entropie par exemple)
			float branchFact=0.0;
			if (valeursAtt.size()-nbTotalExemples>0) {
				proportion=(float)(valeursAtt.size()-nbTotalExemples)/(valeursAtt.size()+countMissingValues);
				branchFact-=proportion*log(proportion);
			}
			if (nbTotalExemples>0) {
				proportion=(float)nbTotalExemples/(valeursAtt.size()+countMissingValues);
				branchFact-=proportion*log(proportion);
			}
			float splitVal;
			if (splitCriteria!='m') splitVal=node->splitValue;
			else splitVal=1.0; // car trop souvent la valeur recomposee est tres proche de celle du pere
			float tmpVal=(((float)(valeursAtt.size()-nbTotalExemples)/(valeursAtt.size()+countMissingValues))*splitGauche+((float)nbTotalExemples/(valeursAtt.size()+countMissingValues))*splitDroite);

			// on poursuit seulement si amelioration par rapport au split indice du pere
			if (tmpVal<node->splitValue) {
				splitVal-=tmpVal;
				if (branchFact>0.0) splitVal/=branchFact;
				else splitVal-=1.0; // on ne veut pas des attributs non discriminants!
				if (splitVal>bestSplitVal) {
					bestSplitVal=splitVal;
					cutFloat=f;
					indexAtt=i;
				}
			}
		}
	}

	delete [] proportions;
	return bestSplitVal;
}
float Classif::findBestSplitOrdered(Noeud* node,int& cutInt,int& indexAtt) {
	if (attDiscOrdonnes.empty()) return -1.0;
	// on commence par trier les attributs par valeurs croissantes,
	// puis on teste de l'ordre de sqrt(n) splits differents equirepartis
	// on garde le meilleur (tous attributs continus confondus), tenant compte des missing_values
	float bestSplitVal=-1.0;
	vector<IntAndClass> valeursAtt;
	float* proportions=new float[nbClasses];
	for (int i=attDiscOrdonnes.size()-1; i>=0; i--) {
		// tentative pour l'attribut discret numero i
		
		// a) recuperation des valeurs des attributs
		int countMissingValues=0;
		valeursAtt.clear();
		for (int j=node->distribExemples.size()-1; j>=0; j--) {
			for (int k=node->distribExemples[j].occurrences.size()-1; k>=0; k--) {
				if (!attDiscOrdonnes[i][node->distribExemples[j].occurrences[k]].miss)
					valeursAtt.push_back(IntAndClass(attDiscOrdonnes[i][node->distribExemples[j].occurrences[k]].val,node->distribExemples[j].index));
				else countMissingValues++;
			}
		}
		// tri des valeurs
		sort(valeursAtt.begin(),valeursAtt.end());

		// b) tests de splits
		float range=valeursAtt.back().val-valeursAtt[0].val; // amplitude de variation de l'attribut
		float step=max(1.0,range/sqrt((float)valeursAtt.size())); // ecart entre deux tests de splits
		float maxProp,proportion;
		for (float f=valeursAtt[0].val+step/2.0; f<=valeursAtt.back().val; f+=step) {

			// calcul du split indice a gauche: on cherche d'abord les proportions
			for (int j=nbClasses-1; j>=0; j--) proportions[j]=0.0; // remise a zero des proportions
			int nbTotalExemples=0,k;
			for (k=0;;k++) {
				if (valeursAtt[k].val>f) break;
				nbTotalExemples++;
				proportions[valeursAtt[k].classIndex]+=1.0;
			}
			float splitGauche;
			if (nbTotalExemples==0) splitGauche=0.0; // attention aux divisions par 0!
			else {
				switch (splitCriteria) {
					case 'g':
						splitGauche=1.0;
						for (int j=nbClasses-1; j>=0; j--) splitGauche-=Noeud::square(proportions[j]/nbTotalExemples);
						break;
					case 'e':
						splitGauche=0.0;
						for (int j=nbClasses-1; j>=0; j--) {
							proportion=proportions[j]/nbTotalExemples;
							if (proportion>0) splitGauche-=proportion*log(proportion);
						}
						break;
					case 'm':
						splitGauche=1.0;
						maxProp=0.0;
						for (int j=nbClasses-1; j>=0; j--) {
							proportion=proportions[j]/nbTotalExemples;
							if (proportion>maxProp) maxProp=proportion;
						}
						splitGauche-=maxProp;
						break;
					default: 
						splitGauche=0.0; // jamais atteint mais evite les warnings
						break;
				}
			}

			// calcul du split indice a droite: on cherche d'abord les proportions
			for (int j=nbClasses-1; j>=0; j--) proportions[j]=0.0; // remise a zero des proportions
			nbTotalExemples=0;
			for (;k<(int)valeursAtt.size();k++) {
				nbTotalExemples++;
				proportions[valeursAtt[k].classIndex]+=1.0;
			}
			float splitDroite;
			if (nbTotalExemples==0) splitDroite=0.0; // attention aux divisions par 0!
			else {
				switch (splitCriteria) {
					case 'g':
						splitDroite=1.0;
						for (int j=nbClasses-1; j>=0; j--) splitDroite-=Noeud::square(proportions[j]/nbTotalExemples);
						break;
					case 'e':
						splitDroite=0.0;
						for (int j=nbClasses-1; j>=0; j--) {
							proportion=proportions[j]/nbTotalExemples;
							if (proportion>0) splitDroite-=proportion*log(proportion);
						}
						break;
					case 'm':
						splitDroite=1.0;
						maxProp=0.0;
						for (int j=nbClasses-1; j>=0; j--) {
							proportion=proportions[j]/nbTotalExemples;
							if (proportion>maxProp) maxProp=proportion;
						}
						splitDroite-=maxProp;
						break;
					default: 
						splitDroite=0.0; // jamais atteint mais evite les warnings
						break;
				}
			}

			// on peut alors donner le split indice pour la valeur f:
			// normalisation (ameliore les perfs de l'entropie par exemple)
			float branchFact=0.0;
			if (valeursAtt.size()-nbTotalExemples>0) {
				proportion=(float)(valeursAtt.size()-nbTotalExemples)/(valeursAtt.size()+countMissingValues);
				branchFact-=proportion*log(proportion);
			}
			if (nbTotalExemples>0) {
				proportion=(float)nbTotalExemples/(valeursAtt.size()+countMissingValues);
				branchFact-=proportion*log(proportion);
			}
			float splitVal;
			if (splitCriteria!='m') splitVal=node->splitValue;
			else splitVal=1.0; // car trop souvent la valeur recomposee est tres proche de celle du pere
			float tmpVal=(((float)(valeursAtt.size()-nbTotalExemples)/(valeursAtt.size()+countMissingValues))*splitGauche+((float)nbTotalExemples/(valeursAtt.size()+countMissingValues))*splitDroite);
			
			// on poursuit seulement si amelioration par rapport au split indice du pere
			if (tmpVal<node->splitValue) {
				splitVal-=tmpVal;
				if (branchFact>0.0) splitVal/=branchFact;
				else splitVal-=1.0; // on ne veut pas des attributs non discriminants!
				if (splitVal>bestSplitVal) {
					bestSplitVal=splitVal;
					cutInt=(int)f;
					indexAtt=i;
				}
			}
		}
	}

	delete [] proportions;
	return bestSplitVal;
}

// fonctions pour determiner ou va aller une valeur manquante
int Classif::findIndexDiscrete(Noeud* node,const vector<VectorString>& subsets,int indexAtt,const string& str) {
	for (int i=node->distribExemples.size()-1; i>=0; i--) {
		if (node->distribExemples[i].str==str) {
			// on initialise les repartitions des classes
			int* repartitions=(int*)calloc(subsets.size(),sizeof(int));	
			int index,sumRep=0,randRes;
			for (int j=node->distribExemples[i].occurrences.size()-1; j>=0; j--) {
				if (!attDiscrets[indexAtt][node->distribExemples[i].occurrences[j]].miss) {
					index=getIndexIn(subsets,attDiscrets[indexAtt][node->distribExemples[i].occurrences[j]].str);
					if (index>=0) {
						repartitions[index]++;
						sumRep++;
					}
				}
			}
			for (int j=subsets.size()-1; j>=0; j--) {
				if (repartitions[j]>=Arbre::NBMIN_REPDISC) {
					randRes=rand()%sumRep+1;
					sumRep=repartitions[0];
					for (int k=0; k<(int)subsets.size(); k++) {
						if (randRes<=sumRep) return k;
						sumRep+=repartitions[k+1];
					}
				}
			}
			free(repartitions);
			break;
		}
	}
	// cas ou on ne trouve pas assez d'elements significatifs
	/*** extension future:
	 *** definir une metrique sur l'espace des attributs,
	 *** et chercher les classes des exemples les plus proches au sens de cette metrique. */
	return -1;
}
int Classif::findIndexContinue(Noeud* node,float cutFloat,int indexAtt,const string& str) {
	for (int i=node->distribExemples.size()-1; i>=0; i--) {
		if (node->distribExemples[i].str==str) {
			int aGauche=0,aDroite=0;
			for (int j=node->distribExemples[i].occurrences.size()-1; j>=0; j--) {
				if (!attContinus[indexAtt][node->distribExemples[i].occurrences[j]].miss) {
					if (attContinus[indexAtt][node->distribExemples[i].occurrences[j]].val<=cutFloat) aGauche++;
					else aDroite++;
				}
			}
			if (aGauche>=Arbre::NBMIN_REPBIN || aDroite>=Arbre::NBMIN_REPBIN) {
				if (rand()%(aGauche+aDroite)+1<=aGauche) return 0;
				else return 1;
			}
			break;
		}
	}
	// cas ou on ne trouve pas assez d'elements significatifs
	/*** extension future:
	 *** definir une metrique sur l'espace des attributs,
	 *** et chercher les classes des exemples les plus proches au sens de cette metrique. */
	return -1;
}
int Classif::findIndexOrdered(Noeud* node,int cutInt,int indexAtt,const string& str) {
	for (int i=node->distribExemples.size()-1; i>=0; i--) {
		if (node->distribExemples[i].str==str) {
			int aGauche=0,aDroite=0;
			for (int j=node->distribExemples[i].occurrences.size()-1; j>=0; j--) {
				if (!attDiscOrdonnes[indexAtt][node->distribExemples[i].occurrences[j]].miss) {
					if (attDiscOrdonnes[indexAtt][node->distribExemples[i].occurrences[j]].val<=cutInt) aGauche++;
					else aDroite++;
				}
			}
			if (aGauche>=Arbre::NBMIN_REPBIN || aDroite>=Arbre::NBMIN_REPBIN) {
				if (rand()%(aGauche+aDroite)+1<=aGauche) return 0;
				else return 1;
			}
			break;
		}
	}
	// cas ou on ne trouve pas assez d'elements significatifs
	/*** extension future:
	 *** definir une metrique sur l'espace des attributs,
	 *** et chercher les classes des exemples les plus proches au sens de cette metrique. */
	return -1;
}
