#ifndef _Classif
#define _Classif

#include <string>
#include <vector>
#include <limits>
#include <cmath>
#include <ctime>
#include <cstdlib>
#include <algorithm>
#include <fstream>
#include "Noeud.h"
#include "Arbre.h"

using namespace std;

// types unions pour prendre en compte les missing values

struct StringOuMiss {
	StringOuMiss () {
		miss=false;
	}
	string str;
	bool miss;
};
struct FloatOuMiss {
	FloatOuMiss () {
		miss=false;
		isInit=false;
	}
	float val;
	bool miss;
	bool isInit; // utile pour ajuster le nombre de lignes
};
struct IntOuMiss {
	IntOuMiss () {
		miss=false;
		isInit=false;
	}
	int val;
	bool miss;
	bool isInit; // utile uniquement pour les tests
};

// types utiles pour reordonner les attributs entiers et flottants

struct FloatAndClass {
	FloatAndClass (float v,int clInd) {
		val=v;
		classIndex=clInd;
	}
	bool operator<(const FloatAndClass& fAndC) const {
		return val<fAndC.val;
	}
	float val;
	int classIndex;
};
struct IntAndClass {
	IntAndClass (int v,int clInd) {
		val=v;
		classIndex=clInd;
	}
	bool operator<(const IntAndClass& iAndC) const {
		return val<iAndC.val;
	}
	int val;
	int classIndex;
};

// type pour une chaine + vecteur booleen utile pour le split symbolique

struct StringAndClassesID {
	StringAndClassesID (const string& s,unsigned long long clID,bool* pb) {
		str=s;
		classesID=clID;
		pbool=pb;
	}
	string str;
	unsigned long long classesID;
	bool* pbool;
};

// la classe principale de creation de l'arbre par discriminations des donnees

class Classif {

public:
	
	Classif () {
		srand(time(NULL)); // pour distribuer les exemples aux arguments manquants
		puiss2=0;
		classes=0;
	}
	
	~Classif () {
		delete [] puiss2;
		delete [] classes;
	}

	// fonctions max auxiliaire
	char max(float valSplitDisc,float valSplitCont,float valSplitOrd) {
		char maximum;
		if (valSplitDisc>valSplitCont) {
			if (valSplitDisc>valSplitOrd) maximum='d';
			else maximum='o';
		}
		else {
			if (valSplitCont>valSplitOrd) maximum='c';
			else maximum='o';
		}
		return maximum;
	}
	
	float max(float a,float b) {
		if (a>b) return a;
		else return b;
	}

	// une fonction auxiliaire qui trouve l'index d'une chaine dans un vecteur
	// declaree static pour pouvoir etre utilisee dans Test
	static int getIndexIn(const vector<VectorString>& subsets,const string& str) {
		for (int i=subsets.size()-1; i>=0; i--) {
			for (int j=subsets[i].size()-1; j>=0; j--)
				if (subsets[i][j]==str) return i;
		}
		return -1; // n'arrive jamais si pas de bugs
	}
	
	// une autre fonction qui indique si une chaine est dans un ensemble de chaines
	bool isIn(const string& str,const VectorString& vs) {
		for (int i=vs.size()-1; i>=0; i--)
			if (vs[i]==str) return true;
		return false;
	}
		
	// fonctions qui convertissent des strings en entiers/flottants via atoi/atof
	// static car utilisees dans Test
	static float toFloat(const string& str) {
		char* ch=(char*)calloc(str.size()+1,sizeof(char));
		for (int i=str.size()-1; i>=0; i--) ch[i]=str[i];
		float res=atof(ch);
		free(ch);
		return res;
	}
	static int toInt(const string& str) {
		char* ch=(char*)calloc(str.size()+1,sizeof(char));
		for (int i=str.size()-1; i>=0; i--) ch[i]=str[i];
		int res=atoi(ch);
		free(ch);
		return res;
	}
	
	// chargement des donnees
	void loadData(char* name,const string& typesAttributs);
	
	// la fonction de creation de l'arbre de classification
	void buildTree(char spC);
	
	// des fonctions auxiliaires qui retournent la meilleure valeur de split
	float findBestSplitDiscrete(Noeud* node,vector<VectorString>& subsets,int& indexAtt);
	float findBestSplitContinue(Noeud* node,float& cutFloat,int& indexAtt);
	float findBestSplitOrdered(Noeud* node,int& cutInt,int& indexAtt);
	
	// fonctions pour determiner ou va aller une valeur manquante
	int findIndexDiscrete(Noeud* node,const vector<VectorString>& subsets,int indexAtt,const string& str);
	int findIndexContinue(Noeud* node,float cutFloat,int indexAtt,const string& str);
	int findIndexOrdered(Noeud* node,int cutInt,int indexAtt,const string& str);
	
	// les attributs ranges par types:
	vector<StringOuMiss*> attDiscrets;
	vector<FloatOuMiss*> attContinus;
	vector<IntOuMiss*> attDiscOrdonnes;
	int nbClasses; // nombre total de classes
	string* classes; // les chaines decrivant les classes

	char* name; // nom du fichier contenant les donnees
	string typesAttributs; // chaine des types d'attributs
	char splitCriteria; // critere de splitting pour la construction de l'arbre
	unsigned long long* puiss2; // tableau des puissances de 2 utiles au split symbolique
	int nbLines; // nombre d'exemples
	Arbre abr; // l'arbre de decision construit
	
};

#endif
