#include "Test.h"

// fichier de sortie contenant les taux d'erreur
ofstream reportErrors("out/errors.txt");

// chargement des donnees de test
void Test::loadData(char* name,const string& typesAttributs) {
	ifstream countLinesFile(name);
	if (!countLinesFile.is_open()) {
		cout << "erreur, le fichier " << name << " n'existe pas" << endl;
		exit(1);
	}
	// on cherche d'abord le nombre de lignes de la BDD test
	nbLines=0; 
	while (countLinesFile.ignore(numeric_limits<int>::max(),'\n')) {
		nbLines++;
	}
	countLinesFile.close();

	// initialisation des structures de donnees (une colonne~=un pointeur)
	// index des classes dans chaque sous-tableau d'attributs (discrets, continus..etc):
	int* tabIndex=new int[typesAttributs.size()]; // indique les index des types d'attributs dans leurs tableaux respectifs
	int nbDisc=0,nbCont=0,nbOrd=0; // pour le remplissage du tableau d'index
	for (int i=0; i<=(int)typesAttributs.size(); i++) {
		if (typesAttributs[i]=='d') {
			attDiscrets.push_back(new StringOuMiss[nbLines]);
			tabIndex[i]=nbDisc;
			nbDisc++;
		}
		else if (typesAttributs[i]=='c') {
			attContinus.push_back(new FloatOuMiss[nbLines]);
			tabIndex[i]=nbCont;
			nbCont++;
		}
		else if (typesAttributs[i]=='o') {
			attDiscOrdonnes.push_back(new IntOuMiss[nbLines]);
			tabIndex[i]=nbOrd;
			nbOrd++;
		}
	}

	// on recopie ensuite la base de donnees dans une structure plus adequate
	ifstream testFile(name);
	string line,strTmp;
	classes=new string[nbLines];
	int countLines=0,countAtt;
	string::iterator curChar,startPoint,memCurChar;
	bool newSubstring; // indique si on est en train de lire une nouvelle valeur
	while (getline(testFile,line,'\n')) {
		if (line.empty()) break; // cas d'un retour a la ligne (?)
		countAtt=0; // compte le nombre d'attributs lus
		newSubstring=true;
		for (curChar=line.begin(); curChar!=line.end(); curChar++) {
			if (newSubstring && *curChar!=',') {
				startPoint=curChar;
				newSubstring=false;
			}
			else if (*curChar==',') { // on ignore les espaces au debut et a la fin des chaines
				while (*startPoint==' ') startPoint++;
				memCurChar=curChar;
				while (*(curChar-1)==' ') curChar--;
				strTmp=string(startPoint,curChar);
				curChar=memCurChar;
				if (strTmp=="?") {
					if (typesAttributs[countAtt]=='d')
						attDiscrets[tabIndex[countAtt]][countLines].miss=true;
					else if (typesAttributs[countAtt]=='c') {
						attContinus[tabIndex[countAtt]][countLines].miss=true;
						attContinus[tabIndex[countAtt]][countLines].isInit=true;
					}
					else if (typesAttributs[countAtt]=='o') {
						attDiscOrdonnes[tabIndex[countAtt]][countLines].miss=true;
						attDiscOrdonnes[tabIndex[countAtt]][countLines].isInit=true;
					}
				}
				else {
					if (typesAttributs[countAtt]=='d')
						attDiscrets[tabIndex[countAtt]][countLines].str=strTmp;
					else if (typesAttributs[countAtt]=='c') {
						attContinus[tabIndex[countAtt]][countLines].val=Classif::toFloat(strTmp);
						attContinus[tabIndex[countAtt]][countLines].isInit=true;
					}
					else if (typesAttributs[countAtt]=='o') {
						attDiscOrdonnes[tabIndex[countAtt]][countLines].val=Classif::toInt(strTmp);
						attDiscOrdonnes[tabIndex[countAtt]][countLines].isInit=true;
					}
				}
				newSubstring=true;
				countAtt++;
			}
			if (countAtt==(int)typesAttributs.size()) break;
		}
		startPoint=++curChar;
		while (curChar!=line.end()) curChar++;
		while (*startPoint==' ') startPoint++;
		while (*(curChar-1)==' ') curChar--;
		classes[countLines]=string(startPoint,curChar);
		countLines++;
	}
	testFile.close();
	delete [] tabIndex;
	
	// test pour savoir si nbLines ou nbLines-1 ..etc (si plusieurs sauts de ligne en fin de fichier)
	// on tire parti du fait que les chaines sont initialisees a "" (similaires aux vecteurs)
	int i=nbLines-1;
	if (!attDiscrets.empty()) 
		while (attDiscrets[0][i].str.empty()) i--;
	else if (!attContinus.empty()) 
		while (!attContinus[0][i].isInit) i--;
	else
		while (!attDiscOrdonnes[0][i].isInit) i--;
	nbLines=i+1; // important pour les stats finales
}

// lance le test sur le jeu de donnees souhaite, et renvoit le nombre d'erreurs
int Test::launchTest(bool sideEffects,string dataType) {
	Noeud* curNode;
	int nbMalClasses=0,indeterminables=0,index;
	bool missingValue; // indique si l'attribut que l'on tente de lire est manquant, auquel cas on sort de la boucle;
						// attention: ce peut etre aussi le cas d'un attribut qui a une valeur symbolique inconnue
	
	if (dataType=="test") {
		for (int i=nbLines-1; i>=0; i--) {
			curNode=classifieur.abr.racine;
			missingValue=false;
			// enfin, l'exploration de l'arbre: (on s'arrte si manque la donnee de test)
			while (curNode->nbFils>0) {
				switch (curNode->type) {
					case 'd':
						index=Classif::getIndexIn(curNode->subsets,attDiscrets[curNode->indexAtt][i].str);
						if (attDiscrets[curNode->indexAtt][i].miss || index<0) {
							missingValue=true;
							break;
						}
						// on recherche le sous-ensemble ou se trouve la valeur d'attribut
						curNode=curNode->fils[index];
						break;
					case 'c':
						if (attContinus[curNode->indexAtt][i].miss) {
							missingValue=true;
							break;
						}
						if (attContinus[curNode->indexAtt][i].val<=curNode->cutFloat) curNode=curNode->fils[0];
						else curNode=curNode->fils[1];
						break;
					case 'o':
						if (attDiscOrdonnes[curNode->indexAtt][i].miss) {
							missingValue=true;
							break;
						}
						if (attDiscOrdonnes[curNode->indexAtt][i].val<=curNode->cutInt) curNode=curNode->fils[0];
						else curNode=curNode->fils[1];
						break;
					default:
						curNode=curNode->fils[0]; // n'arrive jamais
						break;
				}
				if (missingValue) break;
			}
			if (curNode->type=='f') {
				// on arrive donc a une feuille (qui peut etre unknown_classe: cas particulier)
				if (curNode->label=="unknown_classe") indeterminables++;
				else if (classes[i]!=curNode->label) nbMalClasses++;
			}
			else {
				// on a du arreter la recherche car valeur manquante, on regarde si on peut raisonnablement classer le noeud
				int max1=0,max2=0,indexBest=0;
				for (int j=curNode->distribExemples.size()-1; j>=0; j--) {
					if (curNode->distribExemples[j].counter>max1) {
						max2=max1;
						max1=curNode->distribExemples[j].counter;
						indexBest=j;
					}
					else if (curNode->distribExemples[j].counter>max2)
						max2=curNode->distribExemples[j].counter;
				}
				if (max1>0 && (float)max2/max1<Arbre::PRUNING_FACTOR) {
					if (classes[i]!=curNode->distribExemples[indexBest].str) nbMalClasses++;
				}
				else indeterminables++;
			}
		}
	}
	else {
		for (int i=classifieur.nbLines-1; i>=0; i--) {
			curNode=classifieur.abr.racine;
			missingValue=false;
			// enfin, l'exploration de l'arbre: (on s'arrte si manque la donnee de test)
			while (curNode->nbFils>0) {
				switch (curNode->type) {
					case 'd':
						index=Classif::getIndexIn(curNode->subsets,classifieur.attDiscrets[curNode->indexAtt][i].str);
						if (classifieur.attDiscrets[curNode->indexAtt][i].miss || index<0) {
							missingValue=true;
							break;
						}
						// on recherche le sous-ensemble ou se trouve la valeur d'attribut
						curNode=curNode->fils[index];
						break;
					case 'c':
						if (classifieur.attContinus[curNode->indexAtt][i].miss) {
							missingValue=true;
							break;
						}
						if (classifieur.attContinus[curNode->indexAtt][i].val<=curNode->cutFloat) curNode=curNode->fils[0];
						else curNode=curNode->fils[1];
						break;
					case 'o':
						if (classifieur.attDiscOrdonnes[curNode->indexAtt][i].miss) {
							missingValue=true;
							break;
						}
						if (classifieur.attDiscOrdonnes[curNode->indexAtt][i].val<=curNode->cutInt) curNode=curNode->fils[0];
						else curNode=curNode->fils[1];
						break;
					default:
						curNode=curNode->fils[0]; // n'arrive jamais
						break;
				}
				if (missingValue) break;
			}
			if (curNode->type=='f') {
				// on arrive donc a une feuille (qui peut etre unknown_classe: cas particulier)
				if (curNode->label=="unknown_classe") indeterminables++;
				else if (classifieur.classes[i]!=curNode->label) nbMalClasses++;
			}
			else {
				// on a du arreter la recherche car valeur manquante, on regarde si on peut raisonnablement classer le noeud
				int max1=0,max2=0,indexBest=0;
				for (int j=curNode->distribExemples.size()-1; j>=0; j--) {
					if (curNode->distribExemples[j].counter>max1) {
						max2=max1;
						max1=curNode->distribExemples[j].counter;
						indexBest=j;
					}
					else if (curNode->distribExemples[j].counter>max2)
						max2=curNode->distribExemples[j].counter;
				}
				if (max1>0 && (float)max2/max1<Arbre::PRUNING_FACTOR) {
					if (classifieur.classes[i]!=curNode->distribExemples[indexBest].str) nbMalClasses++;
				}
				else indeterminables++;
			}
		}
	}

	if (!sideEffects) return nbMalClasses;

	if (dataType=="test") {
		reportErrors << endl << "***** Erreurs sur l'ensemble de test: *****" << endl << endl;
		if (indeterminables==0) reportErrors << "taux d'erreur: " << (float)nbMalClasses/nbLines << endl << endl;
		else {
			reportErrors << "taux d'erreur par rapport aux exemples que l'arbre peut classer: " << (float)nbMalClasses/(nbLines-indeterminables) << endl;
			reportErrors << "taux d'erreur par rapport a tous les exemples: " << (float)nbMalClasses/nbLines << endl << endl;
		}
		reportErrors << "nombre d'erreurs: " << nbMalClasses << endl;
		reportErrors << "nombre total d'exemples: " << nbLines << endl;
		if (indeterminables>0) reportErrors << "nombre d'exemples inclassables (car donnees d'entrainement pas assez representatives): " << indeterminables << endl;
		reportErrors.close();
	}
	else {
		// vient toujours avant l'affichage precedent
		reportErrors << "***** Erreurs sur l'ensemble d'entrainement: *****" << endl << endl;
		if (indeterminables==0) reportErrors << "taux d'erreur: " << (float)nbMalClasses/classifieur.nbLines << endl << endl;
		else {
			reportErrors << "taux d'erreur par rapport aux exemples que l'arbre peut classer: " << (float)nbMalClasses/(classifieur.nbLines-indeterminables) << endl;
			reportErrors << "taux d'erreur par rapport a tous les exemples: " << (float)nbMalClasses/classifieur.nbLines << endl << endl;
		}
		reportErrors << "nombre d'erreurs: " << nbMalClasses << endl;
		reportErrors << "nombre total d'exemples: " << classifieur.nbLines << endl;
		if (indeterminables>0) reportErrors << "nombre d'exemples inclassables (car donnees d'entrainement pas assez representatives): " << indeterminables << endl;
	}
	return nbMalClasses;
}
