#include <string>
#include <fstream>
#include <iostream>
#include "Arbre.h"
#include "Prune.h"
#include "Classif.h"
#include "Test.h"

using namespace std;

int main(int argc, char** argv) {

	// argv[1]: nom du fichier
	// argv[2]: types des attributs, par exemple "idccddo"==identifiant-discret-continu-..
	// pour connaitre le nombre d'attributs, il suffit de lire une ligne 
	// (formatage: att1,..,attn,nom_de_classe; '?' indique un attribut manquant)
	ifstream dataFile(argv[1]);
	string line;
	getline(dataFile,line,'\n');
	dataFile.close();

	// on cherche le nombre d'attributs
	// 3 types d'attribut a gerer differemment: discret non ordonnes,continus,discret ordonnes
	// + type special: identifiant, que l'on ignore car il ne donnera ausune generalisation
	int nbAttributs=0;
	for (string::iterator curChar=line.begin(); curChar!=line.end(); curChar++)
		if (*curChar==',') nbAttributs++;
	string typesAttributs;
	for (int i=0; i<nbAttributs; i++) typesAttributs.push_back(argv[2][i]);
	
	// on passe alors en parametre a la fonction de creation de l'arbre:
	// - le nom du fichier contenant les donnees
	// - la chaine des types de parametres
	// - le type de split desire
	char splitCriteria=argv[3][0];
	Classif classifieur=Classif();
	classifieur.loadData(argv[1],typesAttributs);
	classifieur.buildTree(splitCriteria);
	Test testTree=Test(classifieur);

	// on coupe les branches de l'arbre trop specialisees pour gagner en generalisation
	/*** remarque au niveau de la conception: Test devrait avoir ses membres statiques
	 *** et ne pas etre instanciee, pour plus de coherence dans l'appel suivant;
	 *** en effet testTree contient justement classifieur.abr, mais il serait plus logique
	 *** de supprimer l'argument testTree plutot que classifieur.abr. 
	 *** Cependant dans cette logique Classif, Prune et Arbre devraient subir le meme 
	 *** traitement, ce qui changerait toute la structure du programme.
	 *** Je laisse donc les choses comme ca pour l'instant. */
	if (argc>=5) testTree.loadData(argv[4],typesAttributs);
	pruning(classifieur.abr,testTree,(argc>=5));

	if (argc<5) {
		// test de l'arbre a partir du meme fichier
		testTree.launchTest(true,"training"); // true pour dire qu'on veut des effets de bord dans un fichier
	}
	else {
		// test de l'arbre a partir d'un autre fichier (donnees de test)
		testTree.launchTest(true,"training"); // on regarde aussi l'erreur sur l'ensemble d'entrainement
												// car apres pruning elle peut ne pas etre negligeable
		testTree.launchTest(true,"test");
	}

	// visualisation de l'arbre calcule
	classifieur.abr.visualize();
	
	return 0;
}
